{- | A list where the last element is of another type. -}
module Data.List.Last (T(..), list, last, toTuple, attach, blendConcat, blendMap, blend, squeezeConcat, squeezeMap, squeeze, awareConcat, awareMap, awareFold, break) where
import Prelude (Eq, Show, Ord, Bool(..), error, uncurry)
import Data.Function
import Data.List ((++))
import Control.Monad
import Control.Monad.Writer hiding (Last)
import Control.Applicative
-- import Control.Category.SplitMor
import qualified Data.Tuple.Pair as Pair
import Data.Tuple.M2 (productMor)

{- | glossary: l=last -}
data T a l
	= Cons a (T a l)
	| Last l
	deriving (Eq, Show, Ord)
list :: T t t1 -> [t]
list = blendConcat id (const [])
last :: T t t1 -> t1
last = fix $ \rc xs -> case xs of
	Cons x xs -> rc xs
	Last l -> l
toTuple :: T a l -> ([a], l)
toTuple xs = (list xs, last xs)
attach :: l -> [a] -> T a l
attach l = fix $ \rc xs -> case xs of
	[] -> Last l
	x:xs -> Cons x (rc xs)
{- | As you can see from the isomorphism below, @T@ is not interesting in theory. Its advantage appears in compilation. -}
{-
isomWriter :: Isom (->) (T a l) (Writer [a] l)
isomWriter = SplitMorToIsom (SplitMor {sect = writer . productMor (last, list), retract = uncurry attach . runWriter})
-}
{- | @instance Monad (T a)@ and @instance Monad (Writer [a])@ coincide via the isomorphism above. Obtain instances of @Functor, Applicative@ by @base.Control.Applicative.WrappedMonad@. -}
instance Monad (T a) where
	return = Last
	m >>= f = fix (\rc xs -> case xs of
		Last x -> f x
		Cons x xs -> Cons x (rc xs)
		) m

blendConcat :: (a -> b) -> (l -> [b]) -> T a l -> [b]
blendConcat f lastf = fix $ \rc xs -> case xs of
	Cons x xs -> f x : rc xs
	Last l -> lastf l
blendMap :: (a -> b) -> (l -> b) -> T a l -> [b]
blendMap f lastf = blendConcat f (return . lastf)
blend :: T a a -> [a]
blend = blendMap id id
squeezeConcat :: (b -> a) -> (b -> T a l) -> [b] -> T a l
squeezeConcat f lastf = fix $ \rc xs -> case xs of
	[] -> error "Data.List.Last.squeeze: empty list has no last element"
	[x] -> lastf x
	x:xs -> Cons (f x) (rc xs)
squeezeMap :: (b -> a) -> (b -> l) -> [b] -> T a l
squeezeMap f lastf = squeezeConcat f (Last . lastf)
squeeze :: [a] -> T a a
squeeze = squeezeMap id id
awareConcat :: (a -> b) -> (a -> [b]) -> [a] -> [b]
awareConcat f lastf = fix $ \rc xs -> case xs of
	[] -> []
	[x] -> lastf x
	x:xs -> f x : rc xs
awareMapInternal :: (a -> b) -> (a -> b) -> [a] -> [b]
awareMapInternal f lastf = awareConcat f (return . lastf)
awareMap :: Pair.T (a -> b) -> [a] -> [b]
awareMap (f, lastf) = awareMapInternal f lastf
awareFoldInternal :: Monoid m => (a -> m) -> (a -> m) -> [a] -> m
awareFoldInternal f lastf = fix $ \rc xs -> case xs of
	[] -> mempty
	[x] -> lastf x
	x:xs -> f x `mappend` rc xs
awareFold :: Monoid m => Pair.T (a -> m) -> [a] -> m
awareFold (f, lastf) = awareFoldInternal f lastf

break :: (a -> Bool) -> [a] -> T a [a]
break p = fix $ \rc xs -> case xs of
	[] -> Last []
	x:xss -> case p x of
		False -> Cons x (rc xss)
		True -> Last xs
mapAccum :: (a -> b -> (a, c)) -> a -> [b] -> T c a
mapAccum f = fix $ \rc acc xs -> case xs of
	[] -> Last acc
	x:xs -> case acc `f` x of
		(acc, y) -> Cons y (rc acc xs)

