{- | We see @Pair a@ as

  * an array indexed by @Bool@ with elements of type @a@, where the index @False@ is @0@ and points to the leftmost component of a tuple;

  * a function of type @Bool -> a@.

Many name clashes. A convinient way to import:

@import qualified Data.Tuple.Pair as Pair@ -}
module Data.Tuple.Pair where
import Prelude ()
import Data.Tuple (fst, snd)
import Data.Bool
import Data.Eq
import qualified Data.Function as F
import Data.Tuple.M2

type T a = (a,a)

{- | read an array -}
read (x0, x1) i = if i then x1 else x0

{- | create an array -}
write f = (fmap f) index

{- | @Pair@ is a functor. @fmap f x = write (f  . (read x))@. -}
fmap f (x0, x1) = (f x0, f x1)

{- | See the source. @index = write id@. -}
index = (False, True)

{- | @Pair@ is an instance of the @Applicative@ class -}

{- | @pure = write . const@. -}
pure x = (x, x)

{- @f <*> x = write (\i -> (read f i) (read x i))@. -}
infixl 4 <*>
(<*>) :: T (a -> b) -> T a -> T b
(<*>) = functorMor

liftA = fmap
liftA2 f x0 x1 = pure f <*> x0 <*> x1

{- | See the source. @ap f x = fmap (F.$ x) f@. @ap = flip (fmap . (flip F.$))@. -}
ap :: T (a -> b) -> a -> T b
ap = productMor

{- | See the source. @swap = write . (. not) . read@. -}
swap (x0, x1) = (x1, x0)

{- | @swapIf b p@ swaps @p@ iff @b@. @swapIf = transformIf Pair.swap@. -}
swapIf b p = write (\i -> read p (i /= b))

{- | @updateEndo i f p@ transforms a pointed by @i@ element in @p@ with the endofunction @f@ -}
updateEndo i f p = let pred = (i ==)
	in write (\i -> (if pred i then f else F.id) (read p i))

{- | @update i x p@ updates the element pointed by @i@ in @p@ with the value @x@ -}
update i x = updateEndo i (F.const x)

{- | See the source.
@diagonal = fmap (flip read) index@.
@diagonal <*> x = x@.
@(swap diagonal) <*> x = swap x@. -}
diagonal = (fst, snd)

{- | see the source -}
transpose ((x00, x01), (x10, x11)) = ((x00, x10), (x01, x11))

{- | see the source -}
toList (x0, x1) = [x0, x1]
